    ;MMX_blk_pref_16KB(int len, int num_times_to_repeat, int* x, int* y);
    ; 
    ; TO ASSEMBLE INTO *.obj DO THE FOLLOWING:
    ; ml.exe -coff -c MMX_blk_pref_16KB.asm
    ;
    .586
    .K3D
    .XMM
    _TEXT SEGMENT 
    PUBLIC _MMX_blk_pref_16KB 
    _MMX_blk_pref_16KB PROC NEAR
    ;===================================================================================================
    ;INSTRUCTIONS BELOW SAVE THE REGISTER STATE WITH WHICH THIS ROUTINE WAS ENTERED
    ;    REGISTERS (eax, ecx, edx are considered volatile and assumed to be changed)
    ;    WHILE THE REGISTERS BELOW MUST BE PRESERVED IF THE USER IS CHANGING THEM
    push ebp
    mov ebp,esp
    ;===================================================================================================
    ; parameters passed into routine:
    ; [ebp+8] = len (number of bytes to be copied)
    ; [ebp+12] = num_times_to_repeat
    ; [ebp+16] = ->x
    ; [ebp+20] = ->y
    ;===================================================================================================
    push eax
    push ebx
    push ecx
    push edx
    push esi
    push edi
    ;===================================================================================================
                                ;
    mov edx,DWORD PTR[ebp+12]   ; edx = # of times to perform copy
    mov esi,DWORD PTR[ebp+16]   ; esi -> x
    mov edi,DWORD PTR[ebp+20]   ; edi -> y
    mov ecx,DWORD PTR[ebp+8]    ; ecx -> size
    push ebp                    ; push ebp onto stack
                                ;
    shr ecx,2                   ; ecx = # of double words to copy
    mov ebp,ecx                 ; copy # of floats to copy to ebp 
    shr ebp,12                  ; ebp = number of FULL 16KB blocks in each vector
    mov ebx,4096                ; ebx = the number of floats in 16KB
    imul ebx,ebp                ; ebx = number of floats in entire set of full 16KB blocks
    neg ebx                     ; ebx = - ebx
    add ebx,ecx                 ; ebx = ecx - ebx = # of floats that are remaining at end which
                                ; do not form a full 16KB block
    align 16                    ; align instructions to 16 byte boundary
    ;===================================================================================================
    repeat_measurement:         ; loop over the vector copy by "num_times_to_repeat" times
    ;===================================================================================================
        pushad                     ; preserve all the 32 bit registers on the stack
        or ebp,ebp                 ; if the number of 16KB blocks is less than 1 skip optimized 
        jz bypasssixteenkbcopyloop ; routine for performing "copy" on multiple 16KB blocks 
        mov ecx,4096               ; ecx = number of floats in 16KB 
                                   ;
        emms                       ; empty mmx state (required on K6-2 processors)
        align 16                   ; align instruction address to 16 byte aligned location
        ;===================================================================================================
        ; BLOCK PREFETCHES MULTIPLE 16 KB BLOCKS and PERFORMS "copy" OPERATION ON THEM
        ;===================================================================================================
        sixteenkbcopyloop:             ;
            sub ecx,32                 ; shift the ->x by 2 cachelines for easier to maximize
                                       ;    the space that can be addressed before having to increment
                                       ;    ->x
            mov eax,64                 ; each block prefetch loop fetches 256 bytes or 4 cachelines
                                       ; and in 16KB there are 32 x 256 bytes to prefetch
            align 16                   ; align instruction address to 16 byte aligned location
            ;===============================================================================================
            ; BLOCK PREFETCHING A 16 KB BLOCK OF "x"
            ;===============================================================================================
            xblkprefcopyloop:          ;
                                       ; the following 4 lines fetch "x" cachelines from the end of the
                                        ;     present 16KB block moving backward 16KB
                movq mm0,[esi+ecx*4+64]
                movq mm0,[esi+ecx*4]
                movq mm0,[esi+ecx*4-64]
                movq mm0,[esi+ecx*4-128]
                sub ecx,64             ; subtract from ecx 64 floats fetched 
                dec eax                ; subtract from eax 4 cachelines to be fetched
            jnz xblkprefcopyloop
            ;===============================================================================================
            add ecx,32                 ; add back the 32 bytes to ecx we subtracted above
                                       ;
            mov eax,256                ; below we perform "copy" on 1 cachelines at a time. We just
                                       ; block prefetched 
            align 16                   ; align instruction address to 16 byte aligned location
            ;===============================================================================================
            ; PERFORMING "copy" ON A 16 KB BLOCK OF "x" and "y"
            ;===============================================================================================
            copyloop:                      ; 
                movq mm0,[esi+ecx*4]       ; mmx0 = [x(1),x(0)]
                movntq [edi+ecx*4],mm0     ; [y(1),y(0)] = [x(1),x(0)]
                movq mm1,[esi+ecx*4+8]     ; mmx1 = [x(3),x(2)]
                movntq [edi+ecx*4+8],mm1   ; [y(3),y(2)] = [x(3),x(2)]
                movq mm2,[esi+ecx*4+16]    ; mmx2 = [x(5),x(4)]
                movntq [edi+ecx*4+16],mm2  ; [y(5),y(4)] = [x(5),x(4)]
                movq mm3,[esi+ecx*4+24]    ; mmx3 = [x(7),x(6)]
                movntq [edi+ecx*4+24],mm3  ; [y(7),y(6)] = [x(7),x(6)]
                movq mm4,[esi+ecx*4+32]    ; mmx4 = [x(9),x(8)]
                movntq [edi+ecx*4+32],mm4  ; [y(9),y(8)] = [x(9),x(8)]
                movq mm5,[esi+ecx*4+40]    ; mmx5 = [x(11),x(10)]
                movntq [edi+ecx*4+40],mm5  ; [y(11),y(10)] = [x(11),x(10)]
                movq mm6,[esi+ecx*4+48]    ; mmx6 = [x(13),x(12)]
                movntq [edi+ecx*4+48],mm6  ; [y(13),y(12)] = [x(13),x(12)]
                movq mm7,[esi+ecx*4+56]    ; mmx7 = [x(15),x(14)]
                movntq [edi+ecx*4+56],mm7  ; [y(15),y(14)] = [x(15),x(14)]
                add ecx,16                 ; increment ecx by 1 cacheline
                dec eax                    ; dec # of 128byte blocks to perform "copy" on
            jnz copyloop                   ;
            ;===============================================================================================
            add ecx,4096               ; increment ecx by 4096 to move the addressing of "x" and "y" 
                                       ;     to the next 16KB block
            dec ebp                    ; decrement the number of 16KB blocks to perform "copy" upon
        jnz sixteenkbcopyloop
        ;===================================================================================================
        sub ecx,4096               ; subtract off of ecx the 1024 added above
        imul ecx,4                 ; ecx = 4 * ecx = # of bytes of "x" and "y" copy'd already 
        add esi,ecx                ; ->x incremented to the 1st element of non full 16KB block 
                                   ; to be "copy'd"
        add edi,ecx                ; ->y incremented to the 1st element of non full 16KB block 
                                   ; to be "copy'd"
        ;===================================================================================================
        bypasssixteenkbcopyloop:
        mov eax,ebx                ; ebx = # of floats in non-full 16KB block
        shr eax,4                  ; eax = # of cachelines of floats in non-full 16KB block
        mov ebp,eax                ;
        imul ebp,16                ; ebp = # of floats in full cachelines of non-full 16KB block
        sub ebx,ebp                ; ebx = # of floats in non-full 16KB block that are not in full cachelines
        mov ecx,0                  ; move 0 into ecx 
        cmp eax,0                  ; check to see if the number of full cachelines is 0 and if
        jz copyfinished            ;
                                   ;
        mov ecx,eax                ; ecx = # of full cachelines
        imul ecx,16                ; ecx = # of floats in full cachelines of last 16KB block that's not full
        align 16                   ; align instruction address to 16 byte aligned location
        ;===============================================================================================
        ; BLOCK PREFETCHING REMAINING FULL CACHELINES OF "x"
        ;===============================================================================================
        copyleftoverblkpref:           ;
            movq mm0,QWORD PTR[esi+ecx*4-64]
            sub ecx,16                 ; decrement # of floats to be prefetched 
        jnz copyleftoverblkpref        ;    by 16 or 1 cacheline
        ;===================================================================================================
        align 16                   ; align instruction address to 16 byte aligned location
        ;===============================================================================================
        ; PERFORMING "copy" ON REMAINING FULL CACHELINES OF "x" and "y"
        ;===============================================================================================
        copyleftoverloop:
            movq mm0,QWORD PTR[esi+ecx*4]
            movq mm1,QWORD PTR[esi+ecx*4+8]
            movq mm2,QWORD PTR[esi+ecx*4+16]
            movq mm3,QWORD PTR[esi+ecx*4+24]
            movq mm4,QWORD PTR[esi+ecx*4+32]
            movq mm5,QWORD PTR[esi+ecx*4+40]
            movq mm6,QWORD PTR[esi+ecx*4+48]
            movq mm7,QWORD PTR[esi+ecx*4+56]
            movntq QWORD PTR[edi+ecx*4],mm0
            movntq QWORD PTR[edi+ecx*4+8],mm1
            movntq QWORD PTR[edi+ecx*4+16],mm2
            movntq QWORD PTR[edi+ecx*4+24],mm3
            movntq QWORD PTR[edi+ecx*4+32],mm4
            movntq QWORD PTR[edi+ecx*4+40],mm5
            movntq QWORD PTR[edi+ecx*4+48],mm6
            movntq QWORD PTR[edi+ecx*4+56],mm7
            add ecx,16                   ; increment addressing of "x" and "y" by 1 cacheline
            dec eax                      ; decrement # of cachelines to perform "copy" on by 1
        jnz copyleftoverloop
        ;===============================================================================================
        jmp copyfinished ; jump to condition below to perform "copy" on remaining 
        ; floats 
        ;===============================================================================================
        ; PERFORMING COPY on those elements of "x" that do not lie in full cachelines
        ;===============================================================================================
        copylastloop:
        movq mm0,QWORD PTR[esi+ecx*4]
        movntq QWORD PTR[edi+ecx*4],mm0
        add ecx,1
        dec ebx
        copyfinished:
        or ebx,ebx
        jnz copylastloop
        ;===============================================================================================
        sfence                       ; serialize store operations
        emms                         ; empty mmx state (required on K6-2 processors)
        popad                        ; restore all 32 bit registers from stack 
        dec edx                      ; decrement the # of times to repeat copy test
                                     ;     if eax =0 then the zero flags bit is set. 
    jnz repeat_measurement           ; if the bit is not set then loop back above to 
    ;===============================================================================================
    pop ebp                          ; restore ebp 
                                     ;
    ;===================================================================================================
    ;INSTRUCTIONS BELOW RESTORE THE REGISTER STATE WITH WHICH THIS ROUTINE WAS ENTERED
    ;    REGISTERS (eax, ecx, edx are considered volatile and assumed to be changed)
    ;    WHILE THE REGISTERS BELOW MUST BE PRESERVED IF THE USER IS CHANGING THEM
    pop edi
    pop esi
    pop edx
    pop ecx
    pop ebx
    pop eax
    mov esp,ebp
    pop ebp
    ;=================================================================================================== 
    ;
    ret
    _MMX_blk_pref_16KB ENDP
    _TEXT ENDS
    END

