    ;void MMX_copy(int len, int num_times_to_repeat, int* x, int* y);
    ; 
    ; TO ASSEMBLE INTO *.obj DO THE FOLLOWING:
    ; ml.exe -coff -c MMX_copy.asm
    ;
    .586
    .K3D
    .XMM
    _TEXT SEGMENT 
    PUBLIC _MMX_copy 
    _MMX_copy PROC NEAR
    ;===================================================================================================
    ;INSTRUCTIONS BELOW SAVE THE REGISTER STATE WITH WHICH THIS ROUTINE WAS ENTERED
    ;    REGISTERS (eax, ecx, edx are considered volatile and assumed to be changed)
    ;    WHILE THE REGISTERS BELOW MUST BE PRESERVED IF THE USER IS CHANGING THEM
    push ebp
    mov ebp,esp
    ;===================================================================================================
    ; parameters passed into routine:
    ; [ebp+8] = len (number of bytes to be copied)
    ; [ebp+12] = num_times_to_repeat
    ; [ebp+16] = ->x
    ; [ebp+20] = ->y
    ;===================================================================================================
    push eax
    push ebx
    push ecx
    push edx
    push esi
    push edi
    ;===================================================================================================
    emms                 ; empty mmx state
    mov ecx,[ebp+8]      ; ecx = # of bytes to copy from x[] to y[]
    mov eax,[ebp+12]     ; eax = number of times to perform "rep movsd" copy
    mov esi,[ebp+16]     ; esi = ->"x"
    mov edi,[ebp+20]     ; edi = ->"y"
    add esi,ecx          ; esi ->the end of the last element of x to be copied
    add edi,ecx          ; edi ->the end of the last element of y to be copied
    shr ecx,3            ; divide the number of bytes to copy by 8 to figure
                         ;    out how many quad words are to be copied from
                         ;    x[] to y[]
    neg ecx              ;
    align 16             ; align instructions to 16 byte boundary
    ;===================================================================================================
    repeat_measurement:  ; loop over the vector copy by "num_times_to_repeat" times
        pushad           ; push all the 32 bit registers onto the stack
        ;================================================================================================
        align 16                     ; align instructions to 16 byte boundary
        bytecopyloop:                ; copy 128 bytes of x[] to y[] with each pass through the code below
            movq mm0,[esi+8*ecx]     ; mmx0 = [x(1),x(0)]
            movq mm1,[esi+8*ecx+8]   ; mmx1 = [x(3),x(2)]
            movq mm2,[esi+8*ecx+16]  ; mmx2 = [x(5),x(4)]
            movq mm3,[esi+8*ecx+24]  ; mmx3 = [x(7),x(6)]
            movq [edi+8*ecx],mm0     ; [y(1),y(0)] = [x(1),x(0)]
            movq [edi+8*ecx+8],mm1   ; [y(3),y(2)] = [x(3),x(2)]
            movq [edi+8*ecx+16],mm2  ; [y(5),x(4)] = [x(5),x(4)]
            movq [edi+8*ecx+24],mm3  ; [y(7),y(6)] = [x(7),x(6)]
            movq mm4,[esi+8*ecx+32]  ; mmx4 = [x(9),x(8)]
            movq mm5,[esi+8*ecx+40]  ; mmx5 = [x(11),x(10)]
            movq mm6,[esi+8*ecx+48]  ; mmx6 = [x(13),x(12)]
            movq mm7,[esi+8*ecx+56]  ; mmx7 = [x(15),x(14)]
            movq [edi+8*ecx+32],mm4  ; [y(9),y(8)] = [x(9),x(8)]
            movq [edi+8*ecx+40],mm5  ; [y(11),y(10)] = [x(11),x(10)]
            movq [edi+8*ecx+48],mm6  ; [y(13),y(12)] = [x(13),x(12)]
            movq [edi+8*ecx+56],mm7  ; [y(15),y(14)] = [x(15),x(14)]
            movq mm0,[esi+8*ecx+64]  ; mmx0 = [x(17),x(16)]
            movq mm1,[esi+8*ecx+72]  ; mmx1 = [x(19),x(18)]
            movq mm2,[esi+8*ecx+80]  ; mmx2 = [x(21),x(20)]
            movq mm3,[esi+8*ecx+88]  ; mmx3 = [x(23),x(22)]
            movq [edi+8*ecx+64],mm0  ; [y(17),y(16)] = [x(17),x(16)]
            movq [edi+8*ecx+72],mm1  ; [y(19),y(18)] = [x(19),x(18)]
            movq [edi+8*ecx+80],mm2  ; [y(21),x(20)] = [x(21),x(20)]
            movq [edi+8*ecx+88],mm3  ; [y(23),y(22)] = [x(23),x(22)]
            movq mm4,[esi+8*ecx+96]  ; mmx4 = [x(25),x(24)]
            movq mm5,[esi+8*ecx+104] ; mmx5 = [x(27),x(26)]
            movq mm6,[esi+8*ecx+112] ; mmx6 = [x(29),x(28)]
            movq mm7,[esi+8*ecx+120] ; mmx7 = [x(31),x(30)]
            movq [edi+8*ecx+96],mm4  ; [y(25),y(24)] = [x(25),x(24)]
            movq [edi+8*ecx+104],mm5 ; [y(27),y(26)] = [x(27),x(26)]
            movq [edi+8*ecx+112],mm6 ; [y(29),y(28)] = [x(29),x(28)]
            movq [edi+8*ecx+120],mm7 ; [y(31),y(30)] = [x(31),x(30)]
            add ecx,16               ; shift addressing ahead by 2 cachelines
        js bytecopyloop              ;
        ;===============================================================================================
        popad              ; pop old 32 bit register values off the stack
        dec eax            ; decrement the # of times to repeat copy test
                           ;    if eax =0 then the zero flags bit is set. 
    jnz repeat_measurement ; if the bit is not set then loop back above to 
                           ;    "repeating:" address and repeat instructions between
                           ;    "repeating:" and "jnz repeating"
    ;===================================================================================================
    emms                   ; empty mmx state
    ;===================================================================================================
    ;INSTRUCTIONS BELOW RESTORE THE REGISTER STATE WITH WHICH THIS ROUTINE WAS ENTERED
    ;    REGISTERS (eax, ecx, edx are considered volatile and assumed to be changed)
    ;    WHILE THE REGISTERS BELOW MUST BE PRESERVED IF THE USER IS CHANGING THEM
    pop edi
    pop esi
    pop edx
    pop ecx
    pop ebx
    pop eax
    mov esp,ebp
    pop ebp
    ;=================================================================================================== 
    ;
    ret
    _MMX_copy ENDP
    _TEXT ENDS
    END

