;void SSE_copy(int len, int num_times_to_repeat, int* x, int* y);
; 
; TO ASSEMBLE INTO *.obj DO THE FOLLOWING:
; ml.exe -coff -c SSE_copy.asm
;
.586
.K3D
.XMM
_TEXT SEGMENT 
PUBLIC _SSE_copy 
_SSE_copy PROC NEAR
;===================================================================================================
;INSTRUCTIONS BELOW SAVE THE REGISTER STATE WITH WHICH THIS ROUTINE WAS ENTERED
;    REGISTERS (eax, ecx, edx are considered volatile and assumed to be changed)
;    WHILE THE REGISTERS BELOW MUST BE PRESERVED IF THE USER IS CHANGING THEM
push ebp
mov ebp,esp
;===================================================================================================
; parameters passed into routine:
; [ebp+8] = len (number of bytes to be copied)
; [ebp+12] = num_times_to_repeat
; [ebp+16] = ->x
; [ebp+20] = ->y
;===================================================================================================
push eax
push ebx
push ecx
push edx
push esi
push edi
;===================================================================================================
mov ecx,[ebp+8]     ; ecx = # of bytes to copy from x[] to y[]
mov eax,[ebp+12]    ; eax = number of times to perform "rep movsd" copy
mov esi,[ebp+16]    ; esi = ->"x"
mov edi,[ebp+20]    ; edi = ->"y"
add esi,ecx         ; esi ->the end of the last element of x to be copied
add edi,ecx         ; edi ->the end of the last element of y to be copied
shr ecx,3           ; divide the number of bytes to copy by 8 to figure
                    ;    out how many quad words are to be copied from
                    ;    x[] to y[]
neg ecx             ;
align 16            ; align instructions to 16 byte boundary
;===================================================================================================
repeat_measurement: ; loop over the vector copy by "num_times_to_repeat" times
    pushad          ; push all the 32 bit registers onto the stack
    ;================================================================================================
    align 16                          ; align instructions to 16 byte boundary
    bytecopyloop:                     ; copy 128 bytes of x[] to y[] with each pass through the code below
        movaps xmm0,[esi+8*ecx]       ; xmm0 = [x(3),x(2),x(1),x(0)]
        movaps xmm1,[esi+8*ecx+16]    ; xmm1 = [x(7),x(6),x(5),x(4)]
        movaps xmm2,[esi+8*ecx+32]    ; xmm2 = [x(11),x(10),x(9),x(8)]
        movaps xmm3,[esi+8*ecx+48]    ; xmm3 = [x(15),x(14),x(13),x(12)]
        movntps [edi+8*ecx],xmm0      ; [y(3),y(2),y(1),y(0)] = [x(3),x(2),x(1),x(0)]
        movntps [edi+8*ecx+16],xmm1   ; [y(7),y(6),y(3),y(2)] = [x(7),x(6),x(5),x(4)]
        movntps [edi+8*ecx+32],xmm2   ; [y(11),y(10),y(5),x(4)] = [x(11),x(10),x(9),x(8)]
        movntps [edi+8*ecx+48],xmm3   ; [y(15),y(14),y(7),y(6)] = [x(15),x(14),x(13),x(12)]
        movaps xmm4,[esi+8*ecx+64]    ; xmm4 = [x(19),x(18),x(17),x(16)]
        movaps xmm5,[esi+8*ecx+80]    ; xmm5 = [x(23),x(22),x(21),x(20)]
        movaps xmm6,[esi+8*ecx+96]    ; xmm6 = [x(27),x(26),x(25),x(24)]
        movaps xmm7,[esi+8*ecx+112]   ; xmm7 = [x(31),x(30),x(29),x(28)]
        movntps [edi+8*ecx+64],xmm4   ; [y(19),y(18),y(17),y(16)] = [x(19),x(18),x(17),x(16)]
        movntps [edi+8*ecx+80],xmm5   ; [y(23),y(22),y(21),y(20)] = [x(23),x(22),x(21),x(20)]
        movntps [edi+8*ecx+96],xmm6   ; [y(27),y(26),y(25),y(24)] = [x(27),x(26),x(25),x(24)]
        movntps [edi+8*ecx+112],xmm7  ; [y(31),y(30),y(29),y(28)] = [x(31),x(30),x(29),x(28)]
        add ecx,16                    ;
    js bytecopyloop                   ;
    ;===============================================================================================
    popad              ; pop old 32 bit register values off the stack
    dec eax            ; decrement the # of times to repeat copy test
                       ;    if eax =0 then the zero flags bit is set. 
jnz repeat_measurement ; if the bit is not set then loop back above to 
                       ;    "repeating:" address and repeat instructions between
                       ;    "repeating:" and "jnz repeating" 
;===================================================================================================
;INSTRUCTIONS BELOW RESTORE THE REGISTER STATE WITH WHICH THIS ROUTINE WAS ENTERED
;    REGISTERS (eax, ecx, edx are considered volatile and assumed to be changed)
;    WHILE THE REGISTERS BELOW MUST BE PRESERVED IF THE USER IS CHANGING THEM
pop edi
pop esi
pop edx
pop ecx
pop ebx
pop eax
mov esp,ebp
pop ebp
;=================================================================================================== 
;
ret
_SSE_copy ENDP
_TEXT ENDS
END