;------------------------------------------------------------------------------
; Name: Procedure timer for MASM.
; Version: 0.06
;  Author: Petroizki
;------------------------------------------------------------------------------

	PROCTIMER_RDTSC_LOOPS EQU 8   	   ; defines the loops for setup overhead timing
	PROCTIMER_LOOPS_POWER_OF_2 EQU 5   ; defines the power of the loop count (2^x), i just use this weird loopcount to get rid of divisions (1<=x<=31)
	                                   ;  *BIG WARNING* the data section will grow by the multiply of 8, when loops are added
	PROCTIMER_WARMUP_LOOPS EQU 3       ; non-timed loops of the benchmark code (warmup phase)
	
	__PROCTIMER_LOOPS EQU (1 SHL PROCTIMER_LOOPS_POWER_OF_2)
	
	__proctimer_loop_start proto C :DWORD, :DWORD, :VARARG
	__proctimer_start proto C :DWORD, :DWORD, :VARARG
	
	PTIMER_PROC_COUNT macro address:REQ, arguments:VARARG
		LOCAL arg, count
		
		IFNB <arguments>
			count = 0
			
			FOR arg, <arguments>
				IFDIF <arg>, <>
					count = count + 1
				ENDIF
			ENDM
			
			invoke __proctimer_start, address, count, arguments
		ELSE
			invoke __proctimer_start, address, 0
		ENDIF
	endm
	
	PTIMER_PROC_LOOP_COUNT macro address:REQ, arguments:VARARG
		LOCAL arg, count
		
		IFNB <arguments>
			count = 0
			
			FOR arg, <arguments>
				IFDIF <arg>, <>
					count = count + 1
				ENDIF
			ENDM
			
			invoke __proctimer_loop_start, address, count, arguments
		ELSE
			invoke __proctimer_loop_start, address, 0
		ENDIF
	endm
.code
timer segment
	PROCTIMER_START label dword
	;--------------------------------------------------------------------------
	;--------------------------------------------------------------------------
__proctimer_ret:
	db 0CCh ; position for the ret instruction
	
	;--------------------------------------------------------------------------
	;--------------------------------------------------------------------------
	db 128-(($-PROCTIMER_START) AND 127) dup (0) ; ALIGN 128
__proctimer_benchmark:
	mov ecx, dword ptr [esp + 16]
	test ecx, ecx
	jz __proctimer_no_parameters
	
	lea edx, [esp + 16]
	
@@:	push dword ptr [edx + ecx*4]
	sub ecx, 1
	jnz @B
__proctimer_no_parameters:
	push dword ptr [__PROCTIMER_SAFE_FLAGS]
	xor eax, eax
	cpuid
	rdtsc
	mov dword ptr [__PROCTIMER_TIMING], eax
	mov dword ptr [__PROCTIMER_TIMING + 4], edx
	;; restore registers and flags, as they were before the macro call
	mov eax, dword ptr [__PROCTIMER_SAFE]
	mov ecx, dword ptr [__PROCTIMER_SAFE + 4]
	mov edx, dword ptr [__PROCTIMER_SAFE + 8]
	mov ebx, dword ptr [__PROCTIMER_SAFE + 12]
	popfd
	call dword ptr [__PROCTIMER_OFFSET]
	;; saves the state of the registers
	;; and gets the current time
	mov dword ptr [__PROCTIMER_SAFE + 16], eax
	mov dword ptr [__PROCTIMER_SAFE + 20], ecx
	mov dword ptr [__PROCTIMER_SAFE + 24], edx
	mov dword ptr [__PROCTIMER_SAFE + 28], ebx
	mov eax, 0
	cpuid
	rdtsc
	
	pushfd
	mov ecx, dword ptr [__PROCTIMER_COUNT]
	sub ecx, 1
	mov dword ptr [__PROCTIMER_COUNT], ecx
	pop dword ptr [__PROCTIMER_SAFE_FLAGS + 4]
	
	mov esp, dword ptr [__PROCTIMER_STACK]
	
	;; jump back, if this is just a warmup phase
	cmp ecx, dword ptr __PROCTIMER_LOOPS
	jnb __proctimer_benchmark
	
	sub eax, dword ptr [__PROCTIMER_TIMING]
	sbb edx, dword ptr [__PROCTIMER_TIMING + 4]
	sub eax, dword ptr [__PROCTIMER_INIT_TIME]
	sbb edx, dword ptr [__PROCTIMER_INIT_TIME + 4]
	
	;; save the benchmarked time to the clock count table
	mov dword ptr [__PROCTIMER_TABLE + ecx*8], eax
	mov dword ptr [__PROCTIMER_TABLE + ecx*8 + 4], edx
	
	;; see if we have more loops left
	test ecx, ecx
	jnz __proctimer_benchmark

	retn

	;--------------------------------------------------------------------------
	;--------------------------------------------------------------------------
	db 128-(($-PROCTIMER_START) AND 127) dup (0) ; ALIGN 128
__proctimer_loop_start proc C par0:DWORD, par1:DWORD, par2:VARARG
__proctimer_loop_start endp
	mov dword ptr [__PROCTIMER_COUNT], (__PROCTIMER_LOOPS + PROCTIMER_WARMUP_LOOPS)
	call __proctimer_init
	call __proctimer_results	
__proctimer_end:
	mov ebx, (OFFSET __PROCTIMER_SAFE + 16)

	; restore flags, registers
	push dword ptr [__PROCTIMER_SAFE_FLAGS + 4]
	mov eax, dword ptr [ebx]
	mov ecx, dword ptr [ebx + 4]
	mov edx, dword ptr [ebx + 8]
	mov ebx, dword ptr [ebx + 12]
	popfd
	
	retn
	
	;--------------------------------------------------------------------------
	;--------------------------------------------------------------------------
	db 128-(($-PROCTIMER_START) AND 127) dup (0) ; ALIGN 128
__proctimer_start proc C par0:DWORD, par1:DWORD, par2:VARARG
__proctimer_start endp
	mov dword ptr [__PROCTIMER_COUNT], 1
	call __proctimer_init
	
	mov eax, dword ptr [__PROCTIMER_TABLE]
	mov edx, dword ptr [__PROCTIMER_TABLE + 4]
	
	mov dword ptr [qwPROCTIMER_RESULT], eax
	mov dword ptr [qwPROCTIMER_RESULT + 4], edx
	jmp __proctimer_end
	
	;--------------------------------------------------------------------------
	;--------------------------------------------------------------------------
	db 128-(($-PROCTIMER_START) AND 127) dup (0) ; ALIGN 128
__proctimer_init:
	push edi
	mov edi, (OFFSET __PROCTIMER_SAFE)
	push ebp
	push esi

	; Preserve registers and flags.
	pushfd
	mov dword ptr [edi], eax
	mov dword ptr [edi + 4], ecx
	mov dword ptr [edi + 8], edx
	mov dword ptr [edi + 12], ebx
	pop dword ptr [__PROCTIMER_SAFE_FLAGS]
	
	; set correct ret instruction for the initialization time testing
	; if timed proc has parameters, then 'ret 0' is set (0000C2h), otherwise just 'ret' (C3h)
	mov eax, dword ptr [esp + 24]
	sub eax, 1
	sbb ecx, ecx
	mov edi, PROCTIMER_RDTSC_LOOPS
	not ecx
	or ebp, -1
	add ecx, 0C3h
	or esi, -1
	mov dword ptr [__proctimer_ret], ecx
	
	mov dword ptr [__PROCTIMER_OFFSET], OFFSET __proctimer_ret
	
	call __proctimer_init_loop
	
	mov eax, dword ptr [esp + 20]
	pop esi
	pop ebp
	pop edi
	lea ecx, [esp - 4]
	mov dword ptr [__PROCTIMER_OFFSET], eax
	
	mov dword ptr [__PROCTIMER_STACK], ecx
	call __proctimer_benchmark
	retn
	
	;--------------------------------------------------------------------------
	;--------------------------------------------------------------------------
	db 128-(($-PROCTIMER_START) AND 127) dup (0) ; ALIGN 128
__proctimer_init_loop:
	push dword ptr [__PROCTIMER_SAFE_FLAGS]
	xor eax, eax
	cpuid
	rdtsc
	mov dword ptr [__PROCTIMER_TIMING], eax
	mov dword ptr [__PROCTIMER_TIMING + 4], edx
	mov eax, dword ptr [__PROCTIMER_SAFE]
	mov ecx, dword ptr [__PROCTIMER_SAFE + 4]
	mov edx, dword ptr [__PROCTIMER_SAFE + 8]
	mov ebx, dword ptr [__PROCTIMER_SAFE + 12]
	popfd
	call dword ptr [__PROCTIMER_OFFSET]
	mov dword ptr [__PROCTIMER_SAFE + 16], eax
	mov dword ptr [__PROCTIMER_SAFE + 20], ecx
	mov dword ptr [__PROCTIMER_SAFE + 24], edx
	mov dword ptr [__PROCTIMER_SAFE + 28], ebx
	mov eax, 0
	cpuid
	rdtsc
	
	pushfd
	; subtract the init time
	sub eax, dword ptr [__PROCTIMER_TIMING]
	sbb edx, dword ptr [__PROCTIMER_TIMING + 4]
	pop dword ptr [__PROCTIMER_SAFE_FLAGS + 4]
	
	mov ebx, edx
	cmp eax, ebp
	sbb ebx, esi
	jnb @F
	
	mov ebp, eax
	mov esi, edx
	
@@:	sub edi, 1
	jnz __proctimer_init_loop
	
	mov dword ptr [__PROCTIMER_INIT_TIME], ebp
	mov dword ptr [__PROCTIMER_INIT_TIME + 4], esi
	
	retn
	
	;--------------------------------------------------------------------------
	;--------------------------------------------------------------------------
	db 128-(($-PROCTIMER_START) AND 127) dup (0) ; ALIGN 128
__proctimer_results:
	push ebp
	push edi
	push esi

	mov ecx, __PROCTIMER_LOOPS
	mov ebp, (OFFSET __PROCTIMER_TABLE) - 8
	
	; calculate average of all timings (average = (x1+x2...+x3)/n)
	xor eax, eax
	xor edx, edx
@@:		add eax, dword ptr [ebp + ecx*8]
		adc edx, dword ptr [ebp + ecx*8 + 4]
		sub ecx, 1
	jnz @B
	; divide edx:eax with __PTIMER_LOOPS (= 2^PTIMER_LOOPS_POWER_OF_2)
	mov ebx, edx
	sar edx, PROCTIMER_LOOPS_POWER_OF_2
	shr eax, PROCTIMER_LOOPS_POWER_OF_2
	shl ebx, (32-PROCTIMER_LOOPS_POWER_OF_2)
	or eax, ebx
	; average in edx:eax
	
	; calculate standard deviation (s=sum(abs(x-average)))
	xor edi, edi
	mov ecx, __PROCTIMER_LOOPS
	push edi ; push 0
@@:		mov ebx, dword ptr [__PROCTIMER_TABLE + ecx*8 - 8]
		mov esi, dword ptr [__PROCTIMER_TABLE + ecx*8 - 4]
		sub ebx, eax
		sbb esi, edx
		
		; abs(esi:ebx)
		mov ebp, esi
		sar ebp, 31
		add esi, ebp
		add ebx, ebp
		xor esi, ebp
		xor ebx, ebp
		add esi, ebp
		
		add dword ptr [esp], ebx
		adc edi, esi
		
		mov dword ptr [__PROCTIMER_DIFFERENCE_TABLE + ecx*8 - 8], ebx
		mov dword ptr [__PROCTIMER_DIFFERENCE_TABLE + ecx*8 - 4], esi
	sub ecx, 1
	jnz @B
	pop edx
	mov ebx, edi
	sar edi, PROCTIMER_LOOPS_POWER_OF_2
	shr edx, PROCTIMER_LOOPS_POWER_OF_2
	shl ebx, (32-PROCTIMER_LOOPS_POWER_OF_2)
	or edx, ebx
	; standard deviation in edi:edx
	
	; calculate average again, this time ignoring all the timings that fall beyond standard deviation
	mov ecx, __PROCTIMER_LOOPS
	xor ebx, ebx
	add edx, 1 ; needed for carry flag check
	xor esi, esi
	xor eax, eax
__proctimer_continue:
		mov ebp, dword ptr [__PROCTIMER_DIFFERENCE_TABLE + ecx*8 - 4]
		cmp dword ptr [__PROCTIMER_DIFFERENCE_TABLE + ecx*8 - 8], edx
		sbb ebp, edi
		jnc @F
		
		; this is a good value, add it to the new average count
		add ebx, dword ptr [__PROCTIMER_TABLE + ecx*8 - 8]
		adc eax, dword ptr [__PROCTIMER_TABLE + ecx*8 - 4]
		add esi, 1
		
@@:		sub ecx, 1
		jnz __proctimer_continue
	xor edx, edx
	div esi
	mov dword ptr [qwPROCTIMER_RESULT + 4], eax
	mov eax, ebx
	div esi
	mov dword ptr [qwPROCTIMER_RESULT], eax
	; the final average
	
	pop esi
	pop edi
	pop ebp
	
	retn
	
	db 4096-(($-PROCTIMER_START) AND 4095) dup (0) ; ALIGN 4096
	__PROCTIMER_INIT_TIME dq ?
	__PROCTIMER_TIMING dq ?
	__PROCTIMER_OFFSET dd ?
	__PROCTIMER_SAFE dd 8 dup (?)
	__PROCTIMER_SAFE_FLAGS dd 2 dup (?)
	__PROCTIMER_COUNT dd ?
	__PROCTIMER_STACK dd ?

	db 128-(($-PROCTIMER_START) AND 127) dup (0) ; ALIGN 128
	qwPROCTIMER_RESULT dq ?
	__PROCTIMER_TABLE dq __PROCTIMER_LOOPS dup (?)
	__PROCTIMER_DIFFERENCE_TABLE dq __PROCTIMER_LOOPS dup (?)
	
	; make it fit in 8kb boundary
	db 4096*2-($-PROCTIMER_START) dup (0)
timer ends
